from random import randint as rd
from typing import Union, Iterable, TypeVar

from ..BaseGraphicObject import BaseGraphicObject
from ..Pixel import Pixel
from ...Classes import Buffer, Vec2, Vec3, Vec4, Event, Counter

from ... import Converter

class Widget(BaseGraphicObject):
    _widgets: dict[str, 'Widget'] = {}
    _counter: Counter = Counter()
    
    @classmethod
    def getByName(cls, name: str) -> Union['Widget', None]:
        return cls._widgets.get(name)
    
    @classmethod
    def tryGetByName(cls, name: str) -> tuple[bool, Union['Widget', None]]:
        widget = cls.getByName(name)
        return (
            widget is not None,
            widget
        )
    
    @classmethod
    def removeAll(cls):
        cls._widgets.clear()
    
    @classmethod
    def generateNewWidgetWithName(cls, widget_class: 'Widget', *args, **kwargs) -> 'Widget':
        if not issubclass(widget_class, cls):
            raise ValueError(f'Class "{widget_class.__name__}" is not subclass {cls.__name__}')
        
        class_name = widget_class.__name__
        cls._counter.add(class_name)
        
        kwargs.update({
            "name": f'{class_name}_{cls._counter.get(class_name)}'
        })
        
        return widget_class(*args, **kwargs)
    
    def __init__(
        self,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        super().__init__(
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name,
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        
        if self.name is not None:
            if self.name in self.__class__._widgets:
                raise ValueError(f'Widget name "{self._name}" already used')
            self.__class__._widgets[self.name] = self


from typing import Union, Callable, Iterable

from ...Classes import Vec2, Vec3, Vec4, Anchor, Event
from ..Pixel import Pixel
from .Button import Button
from ... import Converter
from ... import Func

class CheckBox(Button):
    def __init__(
        self, 
        text: str = 'CheckBox', 
        checked_text: Union[str, None] = None,
        text_pixel: Pixel | tuple[Vec3[int] | Iterable[int], Vec3[int] | Iterable[int], str] | str = None, 
        text_anchor: Anchor = Anchor.center, 
        text_max_size: Union[Vec2[int], Iterable[int]] = None,
        checked: bool = False,
        checked_pixel: Pixel | tuple[Vec3[int] | Iterable[int], Vec3[int] | Iterable[int], str] | str = None,
        select_checked_pixel: Pixel | tuple[Vec3[int] | Iterable[int], Vec3[int] | Iterable[int], str] | str = None,
        size: Vec2[int] | Iterable[int] = None, 
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Vec4[int] | Iterable[int] = None, 
        offset_pos: Vec3[int] | Iterable[int] = None, 
        clear_pixel: Pixel | tuple[Vec3[int] | Iterable[int], Vec3[int] | Iterable[int], str] | str = None, 
        name: str | None = None, 
        tabindex: int | None = None, 
        select_clear_pixel: Pixel | tuple[Vec3[int] | Iterable[int], Vec3[int] | Iterable[int], str] | str = None, 
        on_press_functions: Iterable[Callable[[], None]] | Callable[[], None] = [], 
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        super().__init__(
            text=text, 
            text_pixel=text_pixel, 
            text_anchor=text_anchor, 
            text_max_size=text_max_size,
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            tabindex=tabindex, 
            select_clear_pixel=select_clear_pixel, 
            on_press_functions=on_press_functions, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )

        self._checked_text = Converter.toText(checked_text)
        self._change_checked_event = Event()
    
        self._checked: bool = checked
        self._checked_pixel = Converter.toPixel(checked_pixel)
        self._select_checked_pixel = Converter.toPixel(select_checked_pixel)
        
        self.pressed_event.add(self.switchChecked)
        
    def getClearPixel(self) -> Pixel | None:
        return Func.choisePixel(
            (self.select_checked_pixel if self.selected else self.checked_pixel) if self.checked else None, 
            super().getClearPixel()
        )
    
    def getText(self) -> str:
        return self.checked_text if self.checked else super().getText()
    
    def getChecked(self) -> bool:
        return self._checked
    def setChecked(self, value: bool):
        self._checked = value
        self.setFlagRefresh()
    def switchChecked(self):
        self.checked = not self.checked
            
    @property
    def checked(self) -> bool:
        return self.getChecked()
    @checked.setter
    def checked(self, value: bool):
        self.setChecked(value)
    
    @property
    def checked_pixel(self) -> Union[Pixel, None]:
        return self._checked_pixel
    @checked_pixel.setter
    def checked_pixel(self, value: Union[Pixel, None]):
        self._checked_pixel = value
        self.setFlagRefresh()
    
    @property
    def select_checked_pixel(self) -> Union[Pixel, None]:
        return self._select_checked_pixel
    @select_checked_pixel.setter
    def select_checked_pixel(self, value: Union[Pixel, None]):
        self._select_checked_pixel = value
        self.setFlagRefresh()
        
    @property
    def change_checked_event(self) -> Event:
        return self._change_checked_event
    
    @property
    def checked_text(self) -> str:
        return self._checked_text
    @checked_text.setter
    def checked_text(self, value: str):
        self._checked_text = Converter.toText(value)
        self.setFlagRefresh()


import asyncio
import time

from ..Classes import Event, Counter
from ..Log import Log
from ..Config import Config
from .. import Func


class BaseThread:
    _threads: dict[str, 'BaseThread'] = {}
    init_all_event: Event = Event()
    _init_all_event_called: bool = False
    _amt_sim = Counter()
    
    @classmethod
    def stopAll(cls):
        for thread in tuple(cls._threads.values()):
            if thread.is_enable:
                thread.stop()
    
    def __init__(self, delay: float = 0.5):
        self._init_event = Event()
        self._sim_event = Event(self.simulation, error_ignored=False)
        self._term_event = Event(self.termination)
        self._inited = False
        self._termed = False
        self._enabled = True
        self._delay = delay
        
        self.__class__._threads[self.name] = self
        
    async def run(self):
        try:
            await self._init_event.invokeAsync()
            self._inited = True
            Log.writeOk(f'initialized', self)
                        
            self._checkAllInitialized()
                        
            while self._enabled:
                # t1 = time.perf_counter()
                
                if Func.every(f'_sps_{self.name}', self.delay, True):
                    await self._sim_event.invokeAsync()
                    self.__class__._amt_sim.add(self.name)
                
                # sleep_time = self.delay - (time.perf_counter()-t1)
                
                await asyncio.sleep(
                    0
                    # max(
                    #     sleep_time, 
                    #     0
                    # )
                )

        except asyncio.CancelledError:
            if Config.DEBUG_SHOW_CANCELLED_THREAD_MESSAGE:
                Log.writeNotice('cancelled', self)
            
        except:
            Log.writeError(self)
            BaseThread.stopAll()
            
        finally:
            await self._term_event.invokeAsync()
            self._termed = True
            Log.writeOk(f'terminated', self)
            
            self.__class__._threads.pop(self.name)
    
    def _checkAllInitialized(self):
        if self.__class__._init_all_event_called is True or \
            False in (thread.is_init and thread.is_term is False for thread in self.__class__._threads.values()):
            return
        
        self.__class__.init_all_event.invoke()
        self.__class__._init_all_event_called = True
        
        Log.writeOk(f'all threads initialized', self)
        
    def __str__(self):
        return f'{self.__class__.__name__}'
    
    def stop(self):
        self._enabled = False
    
    async def simulation(self):
        pass
    async def termination(self):
        pass
    
    @property
    def name(self) -> str:
        return self.__class__.__name__
    @property
    def is_init(self) -> bool:
        return self._inited
    @property
    def is_term(self) -> bool:
        return self._termed
    @property
    def is_enable(self) -> bool:
        return self._enabled
    @property
    def delay(self) -> float:
        return self._delay
    @delay.setter
    def delay(self, value: float):
        self._delay = value
    @property
    def init_event(self) -> Event:
        return self._init_event
    @property
    def sim_event(self) -> Event:
        return self._sim_event
    @property
    def term_event(self) -> Event:
        return self._term_event


import asyncio
from typing import Callable, Iterable, overload

from ..Log import Log

class Event:
    @overload
    def __init__(self, *args, error_ignored: bool = False): ...
    
    def __init__(self, *args: Callable[[any], None], **kwargs):
        self._funcs: dict[str, Callable[[any], None]] = {}
        for func in args:
            self.add(func)
            
        self._error_ignored = kwargs.get('error_ignored', False)
    
    def dec(self, func, *args, **kwargs):
        '''
            Add func in event
        '''
        self.add(func)
     
    def add(self, *args: Callable[[], None]):
        for func in args:
            if not issubclass(func.__class__, Callable):
                raise ValueError('The resulting function is not such')
            self._funcs[f'{func.__module__}.{func.__name__}_{tuple(func.__annotations__.keys())}'] = func

    def invoke(self, *args, **kwargs):
        functions = tuple(self._funcs.values())
        for func in functions:
            try:
                res = func(*args, **kwargs)
                if asyncio.iscoroutine(res):
                    raise 
                
            except Exception as ex:
                if not self._error_ignored:
                    raise ex
                Log.writeError()
                
                
    async def invokeAsync(self, *args, **kwargs):
        for func in self._funcs.values():
            try:
                res = func(*args, **kwargs)
                if asyncio.iscoroutine(res):
                    await res
                    
            except Exception as ex:
                if not self._error_ignored:
                    raise ex
                Log.writeError()


class EventKwargs(Event):
    def add(self, func: Callable[[any], None]):
        super().add(func)
    
    def invoke(self, **kwargs):
        super().invoke(**kwargs)
                
    async def invokeAsync(self, **kwargs):
        super().invokeAsync(**kwargs)


import asyncio

class Config:
    ASYNC_EVENT_LOOP = asyncio.get_event_loop()
    
    SPS = 0
    FPS = 0
    
    CLEAR_LINES = 45
    
    NEW_LINE_SYMBOL = ' '
    
    
    # Core
    CORE_MODIFY_WIN_REGEDIT = True
    CORE_WRITE_WARNING_DYNAMIC_MODULE = True
    
    # Log
    LOG_ERROR_DELAY = 4
    LOG_WARNING_DELAY = 3

    # Parser
    PARSER_SKIP_UNKNOWED_ANNOTATIONS = True
    
    TAB_LENGTH = 4
    
    # Drawer
    DRAWER_MAX_SIZE_CACHE = 5
    
    # Pixel
    PIXEL_COLOR_LIMIT = False
    PIXEL_RED_DEPTH = 16
    PIXEL_GREEN_DEPTH = 16
    PIXEL_BLUE_DEPTH = 16
    
    # Debug
    DEBUG_SHOW_INPUT_KEY = False
    DEBUG_SHOW_CANCELLED_THREAD_MESSAGE = False
    DEBUG_SHOW_ANSIICOLOR_CHARS = False
    DEBUG_SHOW_DEBUG_DATA = False  
    
    debug_data = {}



from typing import Union, Iterable, Unpack

from ...Classes import Vec2, Vec3, Vec4, Event, Buffer, Keys, Anchor
from .InteractiveWidget import InteractiveWidget
from ..Pixel import Pixel, Pixels
from ... import Converter
from ... import Func
from ..Drawer import Drawer
from ...Config import Config

class TextBox(InteractiveWidget):
    def __init__(
        self, 
        text: str = 'TextBox',
        text_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        text_anchor: Anchor = Anchor.left,
        text_max_size: Union[Vec2[int], Iterable[int]] = None,
        caret_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        tabindex: Union[int, None] = None,
        select_clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        super().__init__(
            text=text,
            text_pixel=text_pixel,
            text_anchor=text_anchor,
            size=size,
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            offset_pos=offset_pos,
            clear_pixel=clear_pixel,
            name=name,
            tab_index=tabindex,
            select_clear_pixel=select_clear_pixel,
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        self._change_caret_event = Event()
        
        self._text_max_size = Converter.toVec2(text_max_size, Vec2(None, None), True)
        self._caret = 0
        self._caret_pixel = Converter.toPixel(caret_pixel, Pixels.black_white)
    
    async def renderTextBuffer(self):
        text_buffer = await super().renderTextBuffer(True)
        
        if text_buffer.width > 0 and text_buffer.height > 0 and self.selected:
            lines = self.text[:self.caret].split('\n')
            
            caret_x, caret_y = len(lines[-1]), len(lines) - 1
            
            text_buffer[caret_x, caret_y] = Pixel.changePixel(
                text_buffer[caret_x, caret_y], 
                front_color=self.caret_pixel.front_color, 
                back_color=self.caret_pixel.back_color
            )
        
        return text_buffer
    
    def pasteText(self, text: chr):
        lines = self.text.split('\n')
        lines_slice = self.text[:self.caret].split('\n')
        line_width = len(lines[len(lines_slice)-1])
        line_height = len(lines)
        
        # добавить проверку text на multiline
        if (self.text_max_size.width is not None and line_width >= self.text_max_size.width and '\n' not in text) or \
            (self.text_max_size.height is not None and line_height >= self.text_max_size.height and '\n' in text):
            return
        
        self.text = self.text[:self.caret] + text + self.text[self.caret:]
        self.caret += len(text)
    
    def deleteSymbol(self, move_caret: bool, count: int = 1):
        if move_caret and self.caret > 0:
            self.caret -= 1
         
        self.text = self.text[:self.caret] + self.text[self.caret + 1:]
        if count > 1:
            self.deleteSymbol(move_caret, count-1)
    
    def inputKey(self, key: str) -> bool:
        match key:
            case Keys.LEFT | Keys.CTRL_LEFT:
                self.caret -= 1
                
                if key == Keys.CTRL_LEFT:
                    while 1 <= self.caret < len(self.text) and self.text[self.caret] == ' ':
                        self.caret -= 1
                        
                    while 1 <= self.caret < len(self.text) and self.text[self.caret] != ' ':
                        self.caret -= 1
                    
            
            case Keys.RIGHT | Keys.CTRL_RIGHT:
                self.caret += 1
                
                if key == Keys.CTRL_RIGHT:
                    while 0 <= self.caret < len(self.text) and self.text[self.caret] == ' ':
                        self.caret += 1
                    
                    while 0 <= self.caret < len(self.text) and self.text[self.caret] != ' ':
                        self.caret += 1
                    
            
            case Keys.BACKSPACE | Keys.CTRL_BACKSPACE:
                self.deleteSymbol(True)
                
            case Keys.DELETE | Keys.CTRL_DELETE:
                self.deleteSymbol(False)
                
            case Keys.ENTER:
                self.pasteText('\n')

            case _:
                if key.isprintable():
                    self.pasteText(key)
                    
                else:
                    return False
        return True

    def getCaret(self) -> int:
        return self._caret
    def setCaret(self, value: int):        
        text_length = len(self.text) + 1
        self._caret = value - value // text_length * text_length
        self._change_caret_event.invoke()
        self.setFlagRefresh()
    @property
    def caret(self) -> int:
        return self.getCaret()
    @caret.setter
    def caret(self, value: int):
        self.setCaret(value)
        
    def getCaretPixel(self) -> Union[Pixel, None]:
        return Func.choisePixel(self._caret_pixel, self.text_pixel, self.clear_pixel) 
    def setCaretPixel(self, value: Union[Pixel, None]):
        self._caret_pixel = value
        
    @property
    def caret_pixel(self) -> Union[Pixel, None]:
        return self.getCaretPixel()
    @caret_pixel.setter
    def caret_pixel(self, value: Union[Pixel, None]):
        self.setCaretPixel(value)
    
    @property
    def change_caret_event(self) -> Event:
        return self._change_caret_event




from typing import Union, Iterable
import sys

from ..Graphic.Pixel import Pixel, Pixels
from ..Graphic.Widgets import Widget
from ..Graphic.Windows import Window
from ..Classes import Buffer, Vec2, Anchor, Orientation
from ..Log import Log
from ..Config import Config
from .KeyboardManager import KeyboardManager
from .. import Func

class WindowManager:
    _window_queue: list[Window] = []
    _index_current_window: int = 0
        
    @classmethod
    def openNewWindow(cls, window: Window, switch_current_window: bool = True):
        if window.name is not None and cls.getByName(window.name) is not None:
            raise ValueError(f'Window name "{window.name}" already used')
        
        cls._window_queue.append(window)
        if switch_current_window:
            cls._index_current_window = len(cls._window_queue) - 1
        
        window.open_event.invoke()
    
    @classmethod
    def closeCurrentWindow(cls):
        cls._window_queue.pop(cls._index_current_window).close_event.invoke()
        if len(cls._window_queue) > 0:
            cls._normalizeIndexCurrentWindow()
    
    @classmethod
    def closeAll(cls, except_names: Iterable[str] = []):
        windows = cls._window_queue[::-1].copy()
        
        for window in windows:
            if window.name in except_names:
                continue
            window.close_event.invoke()
            cls._window_queue.remove(window)
    
    @classmethod
    def getByName(cls, name: str) -> Union[Window, None]:
        for window in cls._window_queue:
            if window.name == name:
                return window
    
    @classmethod
    def getCurrent(cls) -> Union[Window, None]:
        '''
            if count(windows) == 0
                return 'None'
            else
                return `windows[index_current_window]`
        '''
        if len(cls._window_queue) == 0:
            return None
        
        cls._normalizeIndexCurrentWindow()
        
        return cls._window_queue[cls._index_current_window]
    
    @classmethod
    async def render(cls) -> Union[Buffer[Pixel], None]:
        '''
            if count(windows) == 0
                return `None`
            else
                return `Buffer[Pixel]`
        '''
        
        if len(cls._window_queue) == 0:
            return None

        windows: list[tuple[any]] = [
            ((window.offset_pos.x, window.offset_pos.y), await window.render()) for window in sorted(cls._window_queue, key=lambda window: window.offset_z)
        ]
        
        buffer = Buffer(*Func.calculateSizeByItems(cls._window_queue), Pixel.empty)
        
        for window in windows:
            buffer.paste(*window[0], window[1])
        
        return buffer

    @classmethod
    def pressed(cls, key: str, **kwargs):       
        window_current = cls.getCurrent()
        if window_current is None:
            return
        
        # match char:
        #     case '\x00H':
        #         window_current.selectPrevious()
        #     case '\x00P':
        #         window_current.selectNext()
        #     case _:
        window_current.inputKey(key)
        
    @classmethod
    def _normalizeIndexCurrentWindow(cls):
        cls._index_current_window = Func.normalizeIndex(cls._index_current_window, len(cls._window_queue))
    
KeyboardManager.addHandlerToPressedEvent(WindowManager.pressed)



from typing import Union, Iterable

from .Widget import Widget
from ..Pixel import Pixel, Pixels
from ...Classes import Vec2, Vec3, Vec4, Buffer
from ... import Converter


class ProgressBar(Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        percent: float = 0,
        progress_bar_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = Pixels.b_white,
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        '''
            percent: 0-1
        '''
        super().__init__(
            size=size,
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        
        self._percent = percent
        self._progress_bar_pixel = Converter.toPixel(progress_bar_pixel)
    
    async def refresh(self):
        await super().refresh()
        
        self._buffer.paste(
            self.padding.left, 
            self.padding.top,
            Buffer(
                max(min(round(self._size.x * self._percent), self._size.x), 0), 
                self._size.y,
                self._progress_bar_pixel
            )
        )
    
    @property
    def percent(self) -> float:
        '''
            from 0 to 1
        '''
        return self._percent
    @percent.setter
    def percent(self, value: float):
        self._percent = max(min(value, 1), 0)
        self.setFlagRefresh()
        
        




from typing import Union, Iterable

from .Window import Window
from ..Widgets.Widget import Widget
from ..Pixel import Pixel
from ...Classes import Vec3, Vec2, Vec4, Buffer, Anchor, Orientation
from ... import Func
from ... import Converter
from ..Drawer import Drawer


class TitledWindow(Window):
    def __init__(
            self,
            size: Union[Vec2[int], Iterable[int]] = None,
            min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
            max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
            padding: Union[Vec4[int], Iterable[int]] = None,
            auto_size: bool = False,
            offset_pos: Union[Vec3[int], Iterable[int]] = None, 
            clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            name: Union[str, None] = None,
            widgets: Union[Iterable[Widget], Widget] = [], 
            direction: Union[Orientation, None] = Orientation.horizontal,
            gap: int = 0,
            can_be_moved: bool = True,
            frame: bool = False,
            frame_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            title: str = 'unnamed',
            title_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            title_anchor: Union[Anchor, str] = Anchor.center, 
            title_style: int = 0,
            *args, **kwargs
        ):
        '''
            title_style: `int` = 0 | 1
        '''
        
        super().__init__(
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            auto_size=auto_size,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            widgets=widgets,
            direction=direction,
            gap=gap,
            can_be_moved=can_be_moved,
            frame=frame,
            frame_pixel=frame_pixel,
            *args, **kwargs
        )
                
        self._title = title
        self._title_pixel = Converter.toPixel(title_pixel, Pixel((0, 0, 0), (255, 255, 255)))
        self._title_anchor = Converter.toAnchor(title_anchor)
        self._title_buffer: Buffer[Pixel] = Buffer.empty
        self._title_style = title_style
        
        self._flag_renderTitle = True
        
        self.resize_event.add(self.setFlagRenderTitle)
        

    def setFlagRenderTitle(self):
        self._flag_renderTitle = True
        self.setFlagRefresh()
    
    async def renderTitle(self) -> Buffer[Pixel]:        
        match self._title_style:
            case 1:
                buffer = Buffer(
                    self.width + self.padding.horizontal,
                    3,
                    self.clear_pixel
                )
                buffer.paste(
                    0, 0,
                    Drawer.frame(
                        *buffer.size,
                        *self.clear_pixel.getColors()
                    )
                )
                buffer.paste(
                    0, 0,
                    await Drawer.renderTextBuffer(
                        Func.setTextAnchor(
                            self._title,
                            self._title_anchor,
                            max(self.width + self.padding.horizontal - 2, 0),
                            crop=True
                        ),
                        self._title_pixel
                    ),
                    Vec4(1, 1, 1, 1),
                )
            
            case _:
                buffer = Buffer(
                    self.width,
                    1,
                    self.clear_pixel,
                    [
                        Pixel.changePixel(self._title_pixel, part) for part in Func.setTextAnchor(
                            self._title, 
                            self._title_anchor, 
                            self.width + self.padding.horizontal, 
                            crop=True
                        )
                    ]
                )
        
        self._flag_renderTitle = False
        return buffer
    
    async def refresh(self):
        await super().refresh()
        
        if self._flag_renderTitle:
            self._title_buffer = await self.renderTitle()
        
        self._buffer.paste(
            0, 0,
            self._title_buffer,
            func=Drawer.mergeFramePixels
        )
    
    def getPadding(self):
        return super().getPadding() + Vec4(
            2 if self._title_style == 1 else 0, 
            0, 
            0, 
            0
        )
    
    # def renderTitle(self):
    #     match self._title_style:
    #         case 1:
    #             self._title_buffer = Drawer.frame(
    #                 self.width, 3, 
    #                 self._clear_pixel.front_color, 
    #                 self._clear_pixel.back_color
    #             )
    #             title_tex_length = max(self.width - 2, 0)
    #             self._title_buffer.paste(
    #                 1, 1,
    #                 Buffer(
    #                     title_tex_length, 1,
    #                     self._title_pixel,
    #                     [
    #                         Pixel.changePixel(self._title_pixel, part) for part in Func.setTextAnchor(self._title, self._title_anchor, title_tex_length)[:title_tex_length]
    #                     ]
    #                 )
    #             )
            
    #         case _:
    #             self._title_buffer = Buffer(
    #                 self.width, 1, 
    #                 self._title_pixel, 
    #                 [
    #                     Pixel.changePixel(self._title_pixel, part) for part in Func.setTextAnchor(self._title, self._title_anchor, self.width)[:self.width]
    #                 ]
    #             )
        
    #     self._offset_pos_widgets = Vec3(
    #         0, 
    #         max(self._title_buffer.height-1, 0), 
    #         0
    #     )
    #     self.setUpdateAutoSize()
        
    #     self._flag_renderTitle = False

    # def refresh(self):
    #     super().refresh()

    
    # def render(self):
    #     self._offset_pos_widgets = Vec3(0, 2 if self._title_style == 1 else 0, 0)
        
    #     buffer = super().render()
        
    #     if self._flag_renderTitle:
    #         self.renderTitle()
        
    #     buffer.paste(0, 0, self._title_buffer, Drawer.mergeFramePixels)
        
    #     return buffer
    
    @property
    def title(self) -> str:
        return self._title
    @title.setter
    def title(self, value: str):
        self._title = Converter.toText(value)
        self.setFlagRenderTitle()



from .KeyboardManager import KeyboardManager, Keys
from .WindowManager import WindowManager
from .Parser import Parser
from .AnimationManager import AnimationManager
from .SoundManager import SoundManager


from typing import Union
import sys
import os
from time import sleep

from ..Log import Log
from .. import Func
from .WindowManager import WindowManager
from ..Graphic.Widgets import Widget
from ..Graphic.Windows import Window
from ..Classes import Event
from ..Config import Config

from ..Graphic.Windows import *
from ..Graphic.Widgets import *

class Parser:
    _file_path: Union[str, None] = None
    _file_update_time: Union[float, None] = None
    builded_event: Event = Event()
    
    @classmethod
    def setFile(cls, path: str):
        if not os.path.exists(path):
            Log.writeWarning(f'File "{path}" not found', cls)
            return
        
        cls._file_path = path
        
        Log.writeOk(f'File setted', cls)
        
        cls.checkUpdate()
        
    
    @classmethod
    def checkUpdate(cls):
        widgets_module = 'FloriaConsoleGUI.Graphic.Widgets'
        window_module = 'FloriaConsoleGUI.Graphic.Windows'
        temp: dict[str, any] = {}
        
        def parseGraphicObject(data: list[dict[str, any]]) -> list[Union[Window, Widget]]:
            objects: list[Union[Window, Widget]] = []
            for object_data in data:
                object_class = object_data.pop('class')
                
                if object_class == 'temp':
                    temp.update(object_data)
                    continue
                
                object = Func.choiseValue(
                    getattr(sys.modules[widgets_module], object_class, None),
                    getattr(sys.modules[window_module], object_class, None)
                )
                
                if object is None:
                    raise RuntimeError()

                for key, value in tuple(object_data.items()):
                    if Config.PARSER_SKIP_UNKNOWED_ANNOTATIONS and key not in object.__init__.__annotations__:
                        object_data.pop(key)
                        Log.writeNotice(f'widget "{object.__name__}" attribute "{key}" skipped', cls)
                        sleep(1)
                        continue
                    
                    if isinstance(value, str) and value in temp:
                        object_data[key] = temp[value]
                    
                    match key:
                        case 'widgets':
                            object_data[key] = parseGraphicObject(value)
                    
                objects.append(object(**object_data))
            return objects
                
        if cls._file_path is None:
            return
        
        now_file_update_time = os.path.getmtime(cls._file_path)
        
        if now_file_update_time != cls._file_update_time:
            cls._file_update_time = now_file_update_time
            
            WindowManager.closeAll()
            Widget.removeAll()
            
            try:
                for window in parseGraphicObject(Func.readJson(cls._file_path)):
                    WindowManager.openNewWindow(
                        window
                    )
                    
                cls.builded_event.invoke()
                Log.writeOk('windows builded!', cls)
                sleep(1)
            except:
                WindowManager.closeAll()
                Widget.removeAll()
                Log.writeError()


import asyncio

from ..Log import Log
from ..Threads import BaseThread
from ..Config import *
from ..Managers import KeyboardManager as KeyM


class InputThread(BaseThread):
    def __init__(self):
        super().__init__(0.01)
        
    async def simulation(self):
        await asyncio.create_task(asyncio.to_thread(KeyM.simulation))
    
    
    
    


from typing import Union, Callable, overload
from datetime import datetime, date
import json
import time

from .Classes import Anchor, Vec2, Vec3, Buffer
from .Graphic.Pixel import Pixel
from .Graphic.Widgets import Widget
from .Graphic.Windows import Window
from .Graphic.BaseGraphicObject import BaseGraphicObject


def setTextAnchor(text: str, anchor: Anchor, width: Union[int, None] = None, fillchar: chr = ' ', crop: bool = False):
    target_width = width if width is not None else len(text)
    result = '?'
    match anchor:
        case Anchor.center:
            result = text.center(target_width, fillchar)
        case Anchor.right:
            result = text.rjust(target_width, fillchar)
        case _:
            result = text.ljust(target_width, fillchar)
    if crop:
        return result[:width]
    return result

def readFile(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as file:
        return " ".join(file.readlines())

def saveFile(path: str, data: str) -> str:
    with open(path, 'w', encoding='utf-8') as file:
        file.write(data)

def custom_object_hook(data: dict[str, any]):
    for key, value in data.items():
        if isinstance(value, datetime | date):
            data[key] = value.strftime('%d.%m.%Y %H:%M:%S')
        elif isinstance(value, dict):
            data[key] = custom_object_hook(value)
        elif isinstance(value, list | tuple):
            data[key] = list([custom_object_hook(item) for item in value])
    return data

def readJson(path: str, object_pairs_hook: Callable[[tuple[str, any]], tuple[str, any]] = None) -> Union[dict[str, any], list[any]]:
    return json.loads(readFile(path), object_pairs_hook=object_pairs_hook)

def saveJson(path: str, data: dict[str, any]):
    saveFile(path, json.dumps(data, ensure_ascii=False, indent=4))


def calculateSizeByItems(data: list[BaseGraphicObject]) -> tuple[Vec2[int], Vec2[int]]:
    width = height = 0
    for item in data:
        width = max(item.offset_x + item.width + item.padding.horizontal, width)
        height = max(item.offset_y + item.height + item.padding.vertical, height)
    
    return Vec2(width, height)

@overload
def choiseValue(*args: object, default=None) -> Union[object, None]: ...
def choiseValue(*args: object, **kwargs) -> Union[object, None]:
    for arg in args:
        if arg is not None:
            return arg
    return kwargs.get('default')
    
@overload
def choisePixel(*args: Pixel, default=Pixel.empty) -> Union[Pixel, None]: ...
def choisePixel(*args: Pixel, **kwargs) -> Union[Pixel, None]:
    return choiseValue(*args, default=kwargs.get('default', Pixel.empty))


_every_dict: dict[str, list[float, float]] = {}
def every(marker: str, delay: float, first_true: bool = False) -> bool:
    current_time = time.perf_counter()
    marker_data = _every_dict.get(marker)
    
    if marker_data is None:
        _every_dict[marker] = [
            current_time,
            delay
        ]
        return first_true

    elif sum(marker_data) <= current_time:
        _every_dict[marker] = [current_time, delay]
        return True
    return False

def normalizeIndex(index: int, max: int) -> int:
    index -= index // max * max
    if index < 0:
        index += max
    return index





from ..Log import Log
from ..Threads import BaseThread
from ..Config import Config

class SimulationThread(BaseThread):
    def __init__(self):
        super().__init__((1/Config.SPS) if Config.SPS > 0 else 0)
        
    
    


import os, sys
from typing import Union, Iterable

from ..Log import Log
from ..Threads import BaseThread
from ..Managers.WindowManager import WindowManager
from ..Graphic.Pixel import Pixel, Pixels
from ..Graphic.Widgets import Widget
from ..Graphic.Windows import Window
from ..Config import Config
from .. import Func

class GraphicThread(BaseThread):
    def __init__(self):
        super().__init__((1/Config.FPS) if Config.FPS > 0 else 0)
        self._info = {}
    
    async def simulation(self):
        def pixelColorLimit(pixel: Pixel) -> Pixel:
            def _f(arr: Iterable[int]) -> Iterable[int]:
                r, g, b = 255 // Config.PIXEL_RED_DEPTH, 255 // Config.PIXEL_GREEN_DEPTH, 255 // Config.PIXEL_BLUE_DEPTH
                return [
                    ((arr[0]+1) // r * r)-1,
                    ((arr[1]+1) // g * g)-1,
                    ((arr[2]+1) // b * b)-1
                ]
            
            return Pixel(
                _f(pixel.front_color) if pixel.front_color is not None else None,
                _f(pixel.back_color) if pixel.back_color is not None else None,
                pixel.symbol
            ) 
        
        buffer = await WindowManager.render()
        if buffer is None:
            return
        
        buffer_data = [
            pixel if not Config.PIXEL_COLOR_LIMIT else pixelColorLimit(pixel) if pixel is not None else Pixel.empty 
            for pixel in buffer.data
        ]
 
        pixels: list[Pixel] = \
        [
            buffer_data[i].ANSII if i - i // buffer.width * buffer.width == 0 or not Pixel.compareColors(buffer_data[i-1], buffer_data[i]) else buffer_data[i].symbol 
            #buffer_data[i].symbol
            #buffer_data[i].ANSII
            for i in range(len(buffer_data))
        ]
        
        rendered_text = ''.join([
            ''.join(pixels[y*buffer.width : y*buffer.width+buffer.width]) + f'{Pixel.clearANSII}\n' for y in range(buffer.height)
        ])
        
        if Config.DEBUG_SHOW_DEBUG_DATA:
            if Func.every('update_info', 1, True):
                self._info = self.__class__._amt_sim.getAll()
                self.__class__._amt_sim.clearAll()
            
            Config.debug_data.update(self._info)
        
        sys.stdout.write(f'{'\n' * Config.CLEAR_LINES}{rendered_text}{'; '.join([f'{key}={value}' for key, value in Config.debug_data.items()]) if Config.DEBUG_SHOW_DEBUG_DATA else ''}\n')
    


from typing import Union, Callable, Generic, TypeVar, Iterable, overload
from .Event import Event

_T2 = TypeVar('_T2')
class Vec2(Generic[_T2]):
    @overload
    def __init__(self, width: _T2, height: _T2, no_modify=False): ...
    
    @overload
    def __init__(self, x: _T2, y: _T2, no_modify=False): ...
    
    def __init__(self, x: _T2, y: _T2, **kwargs):
        self._x = x
        self._y = y
        
        self._prop_for_iter: tuple[str] = ['_x', '_y']
    
        self._no_modify: bool = kwargs.get('no_modify', False)
    
        self._update_event: Event = Event()
        
        if self._no_modify:
            self._update_event.add(lambda: self.raise_exc(f'{self} cannot be modify'))
        
    @staticmethod
    def raise_exc(message: str):
        raise RuntimeError(message)
    
    def _setValue(self, attrib_name: str, value: _T2):
        self.__setattr__(attrib_name, value)
        self._update_event.invoke()
    
    @property
    def x(self) -> _T2:
        return self._x
    @x.setter
    def x(self, value: _T2):
        self._setValue('_x', value)
    @property
    def width(self) -> _T2:
        return self.x
    @width.setter
    def width(self, value: _T2):
        self.x = value
        
    @property
    def y(self) -> _T2:
        return self._y
    @y.setter
    def y(self, value: _T2):
        self._setValue('_y', value)
    @property
    def height(self) -> _T2:
        return self.y
    @height.setter
    def height(self, value: _T2):
        self.y = value
    
    @property
    def change_event(self) -> Event:
        return self._update_event
        
    def __len__(self) -> int:
        return len(self._prop_for_iter)
        
    # def __getitem__(self, index: int) -> _T2:
    #     return self.__getattribute__(self._prop_for_iter[index])
    
    def __getitem__(self, index: Union[int, tuple[int]]) -> Union[tuple[_T2], _T2]:
        if isinstance(index, int):
            return self.__getattribute__(self._prop_for_iter[index])
        
        return tuple([
            self.__getattribute__(self._prop_for_iter[i]) for i in index
        ])
        
    def __setitem__(self, index: int, value: _T2):
        self.__setattr__(self._prop_for_iter[index], value)
    
    def __iter__(self):
        yield from [self.__getattribute__(attrib_name) for attrib_name in self._prop_for_iter]
    

    def toTuple(self) -> tuple[_T2]:
        return tuple([self.__getattribute__(attrib_name) for attrib_name in self._prop_for_iter])

    @staticmethod
    def _calc(arr1: Iterable, arr2: Iterable, func: Callable[[any, any], any] = lambda x, y: x + y) -> tuple:
        if len(arr1) > len(arr2):
            raise ValueError()
        
        return tuple([func(arr1[i], arr2[i]) for i in range(len(arr1))])
    
    def __add__(self, other: Iterable):
        return self.__class__(*self._calc(self, other, lambda x, y: x + y))
    
    def __sub__(self, other: Iterable):
        return self.__class__(*self._calc(self, other, lambda x, y: x - y))
        
    def __mul__(self, other: Iterable):
        return self.__class__(*self._calc(self, other, lambda x, y: x * y))

    def __truediv__(self, other: Iterable):
        return self.__class__(*self._calc(self, other, lambda x, y: x / y))
    
    def __iadd__(self, other: Iterable):
        data = self + other
        for i in range(len(data)):
            self[i] = data[i]
        return self
    
    def __isub__(self, other: Iterable):
        data = self - other
        for i in range(len(data)):
            self[i] = data[i]
        return self
        
    def __imul__(self, other: Iterable):
        data = self * other
        for i in range(len(data)):
            self[i] = data[i]
        return self

    def __itruediv__(self, other: Iterable):
        data = self / other
        for i in range(len(data)):
            self[i] = data[i]
        return self

    def __eq__(self, value: Iterable):
        return self.toTuple() == value
    
    def __str__(self):
        return f'Vec2({self._x};{self._y})'


_T3 = TypeVar('_T3')
class Vec3(Vec2, Generic[_T3]):
    @overload
    def __init__(self, x: _T2, y: _T2, z: _T3, no_modify=False): ...
    
    def __init__(self, x: _T3, y: _T3, z: _T3, **kwargs):
        super().__init__(x, y, **kwargs)
        self._z = z
        
        self._prop_for_iter = (*self._prop_for_iter, '_z')
    
    def __str__(self):
        return f'Vec3({self._x};{self._y};{self._z})'
    
    @property
    def z(self) -> _T3:
        return self._z
    @z.setter
    def z(self, value: _T3):
        self._setValue('_z', value)

_T4 = TypeVar('_T4')
class Vec4(Vec3, Generic[_T4]):
    @overload
    def __init__(self, top: _T4, bottom: _T4, left: _T2, right: _T2, no_modify=False): ...
    @overload
    def __init__(self, x: _T2, y: _T2, z: _T4, w: _T4, no_modify=False): ...
    
    def __init__(self, x: _T4, y: _T4, z: _T4, w: _T4, **kwargs):
        super().__init__(x, y, z, **kwargs)
        self._w = w
        
        self._prop_for_iter = (*self._prop_for_iter, '_w')
    
    def __str__(self):
        return f'Vec4({self._x};{self._y};{self._z};{self._w})'
    
    @property
    def w(self) -> _T4:
        return self._w
    @w.setter
    def w(self, value: _T4):
        self._setValue('_w', value)
    
    @property
    def top(self) -> _T4:
        return self.x
    @top.setter
    def top(self, value: _T4):
        self.x = value
    
    @property
    def bottom(self) -> _T4:
        return self.y
    @bottom.setter
    def bottom(self, value: _T4):
        self.y = value
    
    @property
    def left(self) -> _T4:
        return self.z
    @left.setter
    def left(self, value: _T4):
        self.z = value
    
    @property
    def right(self) -> _T4:
        return self.w
    @right.setter
    def right(self, value: _T4):
        self.w = value
        
    @property
    def horizontal(self) -> _T4:
        return self.left + self.right
    
    @property
    def vertical(self) -> _T4:
        return self.top + self.bottom
    
    


from typing import Union, Iterable, Unpack

from ...Classes import Vec2, Vec3, Vec4, Event, Buffer, Anchor
from .Widget import Widget
from ..Pixel import Pixel
from ... import Converter
from ... import Func
from ..Drawer import Drawer
from ...Config import Config

class Label(Widget):
    def __init__(
        self, 
        text: str = 'Label',
        text_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        text_anchor: Anchor = Anchor.left,
        text_max_size: Union[Vec2[int], Iterable[int]] = None,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        super().__init__(
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        self._change_text_event = Event()
        
        self._text = Converter.toMultilineText(text)
        self._text_pixel = text_pixel
        self._text_anchor = Converter.toAnchor(text_anchor)
        self._text_max_size = Converter.toVec2(text_max_size, Vec2(None, None), True)
    
    async def renderTextBuffer(self, add_empty_symbol_to_end: bool = False) -> Buffer[Pixel]:
        text_buffer = await Drawer.renderTextBuffer(
            self.text + (' ' if add_empty_symbol_to_end else ''),
            Func.choisePixel(self.text_pixel, self.clear_pixel)
        )
        
        return text_buffer

    async def refresh(self):
        text_buffer = await self.renderTextBuffer()
        
        self.size = text_buffer.size
        
        await super().refresh()
        
        self._buffer.pasteByAnchor(
            0, 0,
            text_buffer,
            self.text_anchor,
            self.padding
        )
    
    def setText(self, value: str):
        lines = [
            line[:(len(line) if self.text_max_size.width is None else self.text_max_size.width)] 
            for line in value.split('\n')
        ]
        self._text = '\n'.join(lines[:(len(lines) if self.text_max_size.height is None else self.text_max_size.height)])
        self._change_text_event.invoke()
        self.setFlagRefresh()
    def getText(self) -> str:
        return self._text
    @property
    def text(self) -> str:
        return self.getText()
    @text.setter
    def text(self, value: str):
        self.setText(value)

    def getTextPixel(self) -> Union[Pixel, None]:
        return self._text_pixel
    def setTextPixel(self, value: Union[Pixel, None]):
        self._text_pixel = value
    @property
    def text_pixel(self) -> Union[Pixel, None]:
        return self.getTextPixel()
    @text_pixel.setter
    def text_pixel(self, value: Union[Pixel, None]):
        self._text_pixel = value
    
    @property
    def text_anchor(self) -> Anchor:
        return self._text_anchor
    @text_anchor.setter
    def text_anchor(self, value: Anchor):
        self._text_anchor = value
    
    @property
    def text_max_size(self) -> Vec2[Union[int, None]]:
        return self._text_max_size
    @text_max_size.setter
    def text_max_size(self, value: Vec2[Union[int, None]]):
        self._text_max_size = value
        
    @property
    def change_text_event(self) -> Event:
        return self._change_text_event


import readchar
from typing import Union, Callable

from ..Config import Config
from ..Classes import Keys
from ..Log import Log

from ..Classes.Event import Event, EventKwargs

class KeyboardManager:
    _events: dict[str, Event] = {}
    _event_binds: dict[str, set[str]] = {}
    
    _pressed_event: EventKwargs = EventKwargs()
    
    @classmethod
    def addHandlerToPressedEvent(cls, func: Callable[[any], None]):
        '''
            func like this: `func(char: chr, **kwargs) -> None`
        '''
        cls._pressed_event.add(func)
    
    @classmethod
    def registerEvent(cls, event_name: str, key: Union[str, None] = None):
        '''
            if `key` is not None then call `KeyboardManager.bindEvent`
        '''
        if event_name in cls._events:
            raise
        cls._events[event_name] = Event()
        
        Log.writeNotice(f'Event "{event_name}" registered')
        
        if key is not None:
            cls.bindEvent(event_name, key)
        
    
    @classmethod
    def bindEvent(cls, event_name: str, key: str):
        if event_name not in cls._events:
            raise
        
        if key not in cls._event_binds:
            cls._event_binds[key] = set()
            
        cls._event_binds[key].add(event_name)
        
    @classmethod
    def bind(cls, event_name: str, func: Callable[[], None]):
        cls._events[event_name].add(func)

    @classmethod
    def simulation(cls):
        key = readchar.readkey()
        if Config.DEBUG_SHOW_INPUT_KEY:
            Log.writeNotice(f'pressed {key}', cls)
        
        for key, event_names in cls._event_binds.items():
            if key == key:
                for event_name in event_names:
                    cls._events[event_name].invoke()

        cls._pressed_event.invoke(key=key)
        
        Config.debug_data['last pressed key'] = key.encode()


from typing import Union, Iterable
import os
import PIL
import PIL.GifImagePlugin
import PIL.Image

from ..Log import Log
from .. import Func
from ..Graphic import Animation, Pixel
from ..Classes import Buffer


class AnimationManager:
    _animations: dict[str, Animation] = {}

    @classmethod
    def get(cls, name: str) -> Animation:
        if name not in cls._animations:
            raise ValueError(
                f'Animation "{name}" not found'
            )
        return cls._animations[name].copy()
    
    @classmethod
    def register(cls, name: str, animation: Animation):
        if name in cls._animations:
            raise ValueError(
                f'Animation with name "{name}" already exists'
            )
        cls._animations[name] = animation
        Log.writeOk(f'Animation "{name}" loaded', cls)
    
    @classmethod
    def load(cls, path: str):
        '''
            Path to a json file with data like:\n
            [
                {
                    name: `str` - animation name,
                    path: `str` - path to img/jpg/jpeg/gif,
                    delay: `float` default=get from source file - delay between frames,
                    loop: `bool` default=True - animation is looped
                }, \n
                ...
            ]
        '''
        dir_path = os.path.dirname(path)
        
        animations_data: list[dict[str, any]] = Func.readJson(path)
        for data in animations_data:
            image = PIL.Image.open(f'./{dir_path}/{data['path']}', formats=['png', 'jpeg', 'gif'])
            
            cls.register(
                data['name'],
                Animation(
                    data['name'],
                    delay=data.get('delay', cls.getDelay(image)),
                    loop=data.get('loop', True),
                    frames=cls.getFrames(image)
                )
            )
    
    @classmethod
    def getDelay(cls, image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> float:
        if isinstance(image, PIL.GifImagePlugin.GifImageFile):
            return image.info.get('duration') / 1000
        return 0
    
    @classmethod
    def getFrames(cls, image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> Iterable[Buffer[Pixel]]:
        def convertImageToBuffer(image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> Buffer[Pixel]:
            return Buffer(
                image.width,
                image.height,
                None,
                [
                    Pixel(back_color=pixel_data[:3]) if pixel_data[3] > 125 else None for pixel_data in image.convert('RGBA').getdata()
                ]
            )
        
        if isinstance(image, PIL.GifImagePlugin.GifImageFile):
            frames = []
            for i in range(image.n_frames):
                image.seek(i)
                frames.append(
                    convertImageToBuffer(image)
                )
            return frames
        
        else:
            return [
                Buffer(
                    image.width,
                    image.height,
                    None,
                    convertImageToBuffer(image)
                )
            ]
                
    


from typing import Union, Iterable
import time
import math

from ..Classes import Vec2, Buffer, Event
from ..Graphic.Pixel import Pixel

class Animation:
    def __init__(
        self, 
        name: str,
        loop: bool = True,
        frames: Union[Iterable[Buffer[Pixel]], Buffer[Pixel]] = (),
        delay: float = 1,
        *args, **kwargs
    ):
        self._name = name
        
        if isinstance(frames, Iterable):
            if len(frames) == 0:
                raise ValueError()
            self._frames: tuple[Buffer[Pixel]] = tuple(frames)
        else:
            self._frames: tuple[Buffer[Pixel]] = (frames)
        
        self._loop = loop
        self._current_frame = 0
        self._delay = delay
        
        self._time = time.perf_counter()
        
        self._end_animation_event = Event()
        

    def render(self) -> Buffer[Pixel]:
        match len(self._frames):
            case 0:
                raise
            case 1:
                return self._frames[0]
            case _:
                if self.loop is False and self.current_frame == len(self._frames) - 1:
                    return self._frames[-1]
                
                if self.is_next:
                    # высчитывает количество "прошедших" кадров
                    self._current_frame += math.floor((time.perf_counter() - self._time)/self.delay)
                    self._time = time.perf_counter()
                    
                    
                    while self._current_frame >= len(self._frames):
                        self._current_frame -= len(self._frames)
                        self._end_animation_event.invoke()
                        
                return self._frames[self.current_frame]

    @property
    def size(self) -> Vec2:
        return self._frames[0].size
    @property
    def width(self) -> int:
        return self.size.width
    @property
    def height(self) -> int:
        return self.size.height
    
    @property
    def is_next(self) -> bool:
        return self._time + self.delay <= time.perf_counter()
    
    @property
    def current_frame(self) -> int:
        return self._current_frame
    
    @property
    def loop(self) -> bool:
        return self._loop
    
    @property
    def delay(self) -> float:
        return self._delay
    @delay.setter
    def delay(self, value: float):
        self._delay = value
        
    @property
    def name(self) -> str:
        return self._name
    
    def __len__(self) -> int:
        return len(self._frames)

    def copy(self) -> 'Animation':
        return Animation(
            self.name,
            self.loop,
            self._frames,
            self.delay
        )


from typing import Union, Iterable, Unpack

from ...Classes import Vec2, Vec3, Vec4, Event, Buffer, Anchor
from .Label import Label
from ..Pixel import Pixel
from ... import Converter

class InteractiveWidget(Label):
    def __init__(
        self, 
        text: str = 'InteractiveWidget',
        text_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        text_anchor: Anchor = Anchor.left,
        text_max_size: Union[Vec2[int], Iterable[int]] = None,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        tabindex: Union[int, None] = None,
        select_clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        super().__init__(
            text=text,
            text_pixel=text_pixel,
            text_anchor=text_anchor,
            text_max_size=text_max_size,
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        self._selected_event = Event()
        self._not_selected_event = Event()
        
        self._select_clear_pixel = Converter.toPixel(select_clear_pixel)
        self._tabindex = tabindex
        self._selected: bool = False
    
    def inputKey(self, key: str) -> bool:
        '''
            return True if the key is processed else False
        '''
        return False
    
    def getClearPixel(self) -> Union[Pixel, None]:
        return self.select_clear_pixel if self.selected else super().getClearPixel()
    
    @property
    def tabindex(self) -> Union[int, None]:
        ''' wip '''
        return self._tabindex
    
    @property
    def selected(self) -> bool:
        return self._selected
    @selected.setter
    def selected(self, value: bool):
        if self._selected != value:
            if value:
                self._selected_event.invoke()
            else:
                self._not_selected_event.invoke()
        
        self._selected = value
        
        self.setFlagRefresh()
    
    @property
    def select_clear_pixel(self) -> Union[Pixel, None]:
        return self._select_clear_pixel
    @select_clear_pixel.setter
    def select_clear_pixel(self, value: Union[Pixel, None]):
        self._select_clear_pixel = value
    
    @property
    def selected_event(self) -> Event:
        return self._selected_event
    
    @property
    def not_selected_event(self) -> Event:
        return self._not_selected_event
    


from typing import Union, Iterable

from ...Classes import Vec2, Vec3, Vec4, Buffer
from .Container import Container
from .Widget import Widget
from ..Pixel import Pixel, Pixels
from ... import Converter

class Scroll(Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None,
        scroll_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        *args, **kwargs
    ):
        super().__init__(
            size=size, 
            padding=padding,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            *args, **kwargs
        )

        self._widgets: list[Widget] = Converter.toListObjects(widgets)

        self._full_buffer: Union[Buffer[Pixel], None] = None
        self._updateFullBuffer()
        self.scroll = Vec2(0, 0)
        self._scroll_pixel = Converter.toPixel(scroll_pixel, Pixel((0, 0, 0), (255, 255, 255)))
        
    def _updateFullBuffer(self):
        max_width = max_height = 0
        
        for widget in self._widgets:
            max_width = max(max_width, widget.size.x + widget.offset_pos.x)
            max_height = max(max_height, widget.size.y + widget.offset_pos.y)

        self._full_buffer = Buffer(max_width, max_height, None)

        for widget in self._widgets:
            self._full_buffer.paste(
                *(widget.offset_pos.toTuple()[:2]),
                widget.render()
            )
    
    async def refresh(self):
        if self._full_buffer is None:
            self._updateFullBuffer()
        
        await super().refresh()
        
        self._buffer.paste(
            -self._scroll.x, 
            -self._scroll.y, 
            self._full_buffer
        )
        
        hor = self._full_buffer.width > self._buffer.width
        ver = self._full_buffer.height > self._buffer.height
        both = hor and ver
        
        if hor:
            for x in range(self._buffer.width):
                self._buffer[
                    x, 
                    self._buffer.height-1
                ] = self._scroll_pixel
                
            if self._scroll.x > 0:
                self._buffer[
                    0,
                    self._buffer.height-1
                ] = Pixel.changePixel(self._scroll_pixel, symbol='◄')
            
            if self._full_buffer.width - (self._scroll.x + self._buffer.width) > 0:
                self._buffer[
                    self._buffer.width-1 - (1 if both else 0), 
                    self._buffer.height-1
                ] = Pixel.changePixel(self._scroll_pixel, symbol='►')
        
        if ver:
            for y in range(self._buffer.height):
                self._buffer[
                    self._buffer.width-1, 
                    y
                ] = self._scroll_pixel
                
            if self._scroll.y > 0:
                self._buffer[
                    self._buffer.width-1, 
                    0
                ] = Pixel.changePixel(self._scroll_pixel, symbol='▲')
                
            if self._full_buffer.height - (self._scroll.y + self._buffer.height) > 0:
                self._buffer[
                    self._buffer.width-1, 
                    self._buffer.height-1 - (1 if both else 0)
                ] = Pixel.changePixel(self._scroll_pixel, symbol='▼')
        
        if hor and ver:
            self._buffer[self._buffer.width-1, self._buffer.height-1] = Pixel.changePixel(self._scroll_pixel, symbol='/')
                
    @property
    def scroll(self) -> Vec2[int]:
        return self._scroll
    @scroll.setter
    def scroll(self, value: Vec2[int]):
        def _valide_scroll():
            self._scroll._x = min(max(self._scroll.x, 0), max(self._full_buffer.width - self.size.x, 0))
            self._scroll._y = min(max(self._scroll.y, 0), max(self._full_buffer.height - self.size.y, 0))
            self.setFlagRefresh()
        
        self._scroll = value
        self._scroll.change_event.add(_valide_scroll)
        self.setFlagRefresh()
    
        


        
        
        
        
        


from typing import Union, Iterable

from ..BaseGraphicObject import BaseGraphicContainerObject
from .Widget import Widget
from ..Pixel import Pixel
from ...Classes import Vec2, Vec3, Vec4, Orientation
from ..Drawer import Drawer
from ... import Converter
from ... import Func

class Container(BaseGraphicContainerObject, Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None,
        direction: Union[Orientation, None] = Orientation.horizontal,
        gap: int = 0,
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        super().__init__( 
            size=size,
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            auto_size=auto_size,
            offset_pos=offset_pos,
            clear_pixel=clear_pixel,
            name=name,
            objects=widgets,
            direction=direction,
            gap=gap,
            can_be_moved=can_be_moved,
            *args, **kwargs
        )



from typing import Union, Iterable

from ...Classes import Vec2, Vec3, Vec4, Event
from .Widget import Widget
from ..Pixel import Pixel
from ..Animation import Animation
from ...Managers.AnimationManager import AnimationManager

class Media(Widget):
    def __init__(
        self,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        animation: Animation = None,
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        
        self._animation: Animation = animation if isinstance(animation, Animation) else AnimationManager.get(animation)
        
        super().__init__(
            size=size if size is not None else self._animation.size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name,
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        
        self._animation_change_event = Event()
        
    
    async def refresh(self):
        await super().refresh()
        
        if self.animation is None:
            return
        
        self._buffer.paste(
            self.padding.left,
            self.padding.top,
            self.animation.render().resize(*self.size)
        )
    
    async def awaitingRefresh(self):
        return self.animation.is_next
    
    async def render(self):
        if self.animation.is_next:
            self.setFlagRefresh()
        
        return await super().render()
    
    def getAnimation(self) -> Animation:
        return self._animation
    def setAnimation(self, value: Union[Animation, None]):
        self._animation = value
        self._animation_change_event.invoke()
        self.setFlagRefresh()
    
    @property
    def animation(self) -> Animation:
        return self.getAnimation()
    @animation.setter
    def animation(self, value: Union[Animation, None]):
        self.setAnimation(value)
    
    @property
    def animation_change_event(self) -> Event:
        return self._animation_change_event
    
        


from typing import Union, Iterable, TypeVar

from .Classes import Vec2, Vec3, Vec4, Anchor, Orientation
from .Graphic.Pixel import Pixel, Pixels
from .Graphic.Animation import Animation
from .Config import Config

def _toVecX(vec_type: type[any], data: Union[any, Iterable], default: any, allow_none: bool = False) -> any:
    if data is None:
        return default
    if not isinstance(data, Iterable) or allow_none is False and None in data:
        raise ValueError()
    return data if isinstance(data, vec_type) else vec_type(*data)

def toVec2(data: Union[Vec2, Iterable], default: Vec2 = Vec2(0, 0), allow_none: bool = False) -> Vec2:
    return _toVecX(Vec2, data, default, allow_none)
def toVec3(data: Union[Vec3, Iterable], default: Vec3 = Vec3(0, 0, 0), allow_none: bool = False) -> Vec3:
    return _toVecX(Vec3, data, default, allow_none)
def toVec4(data: Union[Vec4, Iterable], default: Vec4 = Vec4(0, 0, 0, 0), allow_none: bool = False) -> Vec4:
    return _toVecX(Vec4, data, default, allow_none)

def toPixel(data: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None], default: Pixel = None) -> Union[Pixel, None]:
    '''
        `data` can be of any `Iterable-Type`
    '''
    if data is None:
        return default
    elif isinstance(data, str):
        return Pixels.__dict__[data]
    elif isinstance(data, Pixel | Iterable):
        return data if isinstance(data, Pixel) else Pixel(*data)

    raise ValueError(f'data({data}) is not Pixel | Iterable')

TOLISTOBJECTS_T1 = TypeVar('TOLISTOBJECTS_T1')
def toListObjects(data: Union[Iterable[TOLISTOBJECTS_T1], TOLISTOBJECTS_T1, None]) -> list[TOLISTOBJECTS_T1]:
    if data is None:
        return []
    elif isinstance(data, Iterable):
        return [*data]
    return [data]

def toAnchor(anchor: Union[Anchor, str]) -> Anchor:
    if isinstance(anchor, Anchor):
        return anchor
    elif isinstance(anchor, str):
        return Anchor[anchor]
    raise ValueError()

def toOrientation(orientation: Union[Orientation, str]) -> Orientation:
    if isinstance(orientation, Orientation):
        return orientation
    elif isinstance(orientation, str):
        return Orientation[orientation]
    raise ValueError()

def toText(text: str) -> str:
    return text.replace('\n', '\\n').replace('\t', ' ' * Config.TAB_LENGTH)
def toMultilineText(text: str) -> str:
    return text.replace('\t', ' ' * Config.TAB_LENGTH)


from typing import Union

class Counter:
    def __init__(self):
        self._data: dict[str, int] = {}
    
    def add(self, key: str, amt: int = 1, default_value: int = 0):
        if key not in self._data:
            self._data[key] = default_value
        self._data[key] += amt
    
    def pop(self, key: str, amt: int = 1):
        if key not in self._data:
            raise ValueError(f'Key "{key}" not found')
        self._data[key] -= amt
    
    def get(self, key: str) -> Union[int, None]:
        return self._data.get(key)
    
    def getAll(self) -> dict[str, int]:
        return self._data.copy()
    
    def clearAll(self):
        self._data.clear()


from typing import Union, Iterable, Callable

from ...Classes import Vec2, Vec3, Vec4, Event, Anchor, Keys
from ..Pixel import Pixel
from .InteractiveWidget import InteractiveWidget

class Button(InteractiveWidget):
    def __init__(
        self,
        text: str = 'Button',
        text_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        text_anchor: Anchor = Anchor.center,
        text_max_size: Union[Vec2[int], Iterable[int]] = None,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[int], Iterable[int]] = None,
        max_size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        tabindex: Union[int, None] = None,
        select_clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        on_press_functions: Union[Iterable[Callable[[], None]], Callable[[], None]] = [],
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        super().__init__(
            text=text,
            text_pixel=text_pixel,
            text_anchor=text_anchor,
            text_max_size=text_max_size,
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            tabindex=tabindex, 
            select_clear_pixel=select_clear_pixel, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        self._pressed_event = Event(
            *(on_press_functions if isinstance(on_press_functions, Iterable) else [on_press_functions])
        )
        
    def inputKey(self, key) -> bool:
        match key:
            case Keys.ENTER:
                self.pressed_event.invoke()

            case _:
                return False
        return True
    
    @property
    def pressed_event(self) -> Event:
        return self._pressed_event


UP = '\x00H'
DOWN = '\x00P'
LEFT = '\x00K'
RIGHT = '\x00M'

CTRL_C = '\x03'
CTRL_RIGHT = '\x00t'
CTRL_LEFT = '\x00s'
CTRL_BACKSPACE = '\x7f'
CTRL_DELETE = '\x00\xc2\x93'

ENTER = '\r'
BACKSPACE = '\x08'
DELETE = '\x00S'

SPACE = ' '


from .Buffer import Buffer
from .Event import Event, EventKwargs
from .Vec import Vec2, Vec3, Vec4
from .Anchor import Anchor
from .Orientation import Orientation
from .Counter import Counter
from . import Keys


from typing import Union, Callable, Generic, TypeVar, Iterable
import math

from .Anchor import Anchor
from .Vec import Vec2, Vec4

_T = TypeVar('_T')

class Buffer(Generic[_T]):
    empty = None
    
    def __init__(self, width: int, height: int, defualt_value: _T = None, data: Union[Iterable[_T], None] = None):
        if width < 0 or height < 0:
            raise ValueError(f'Width or height cannot be less than 0')
        
        self._size = Vec2(width, height, no_modify=True)
        
        self._defualt_value = defualt_value
        
        if data is not None:
            if len(data) != width*height:
                raise ValueError(f'The parameters for width({width}) and height({height}) do not match the len of data({len(data)})')
            self._data: list[_T] = list(data)  
        else:
            self.fill()

    @staticmethod
    def pasteFunction(
        x_parent: int, 
        y_parent: int, 
        width_parent: int, 
        height_parent: int, 
        x_child: int, 
        y_child: int, 
        width_child: int, 
        height_child: int, 
        old: Union[any, None], 
        new: Union[any, None]
    ) -> Union[any, None]:
        return new if new is not None else old
    
    def pasteArray(
        self, 
        offset_x: int, 
        offset_y: int, 
        array_width: int, 
        array_height: int, 
        array: Iterable, 
        padding: Vec4 = Vec4(0, 0, 0, 0), 
        func: Callable[[int, int, _T, _T], _T]=None
    ):
        if array is None or array_width == 0 or array_height == 0:
            return
        
        for y in range(
            max(
                -offset_y, 
                0
            ), 
            min(
                array_height - (padding.vertical + array_height + offset_y - self.height),
                array_height
            )):
            for x in range(
                max(
                    -offset_x, 
                    0
                ), 
                min(
                    array_width - (padding.horizontal + array_width + offset_x - self.width),
                    array_width
                )):
                xpos = x + offset_x + padding.left
                ypos = y + offset_y + padding.top
                
                self._data[
                    ypos * self._size.width + xpos
                ] = (func if func is not None else self.pasteFunction)(
                    xpos, ypos,
                    self.width, self.height,
                    x, y,
                    array_width, array_height,
                    self._data[
                        ypos * self._size.width + xpos
                    ],
                    array[
                        y * array_width + x
                    ]
                )
        
        # for y in range(array_height):
        #     for x in range(array_width):
        #         xpos, ypos = x + offset_x + padding.left, y + offset_y + padding.top
                
        #         if xpos >= self.width - padding.right or xpos < 0 + padding.left or\
        #             ypos >= self.height - padding.bottom or ypos < 0 + padding.top:
        #             continue
                
        #         self._data[
        #             ypos * self._size.width + xpos
        #         ] = func(
        #             self._data[
        #                 ypos * self._size.width + xpos
        #             ],
        #             array[
        #                 y * array_width + x
        #             ]
        #         )
    
    def paste(
        self, 
        offset_x: int, 
        offset_y: int, 
        buffer: Union['Buffer', None], 
        padding: Vec4 = Vec4(0, 0, 0, 0), 
        func: Callable[[int, int, _T, _T], _T]=None
    ):
        if buffer is None:
            return
        
        self.pasteArray(
            offset_x, offset_y,
            buffer.width, buffer.height, buffer.data,
            padding,
            func
        )

    def pasteByAnchor(
        self, 
        offset_x: int, 
        offset_y: int, 
        buffer: Union['Buffer', None], 
        anchor: Anchor = Anchor.left_top, 
        padding: Vec4 = Vec4(0, 0, 0, 0), 
        func: Callable[[int, int, _T, _T], _T]=None
    ):
        offset_x_calc = math.floor((self.width - padding.left - padding.right) / 2 - buffer.width/2)
        match anchor:
            case Anchor.left | Anchor.left_top | Anchor.left_bottom:
                offset_x_calc = 0
            case Anchor.right | Anchor.right_top | Anchor.right_bottom:
                offset_x_calc = self.width - buffer.width - padding.left - padding.right
                
        offset_y_calc = math.floor((self.height - padding.top - padding.bottom) / 2 - buffer.height/2)
        match anchor:
            case Anchor.top | Anchor.left_top | Anchor.right_top:
                offset_y_calc = 0
            case Anchor.bottom | Anchor.left_bottom | Anchor.right_bottom:
                offset_y_calc = self.height - buffer.height - padding.top - padding.bottom
        
        self.paste(
            offset_x_calc + offset_x, 
            offset_y_calc + offset_y, 
            buffer,
            padding,
            func
        )
    
    def fill(self, value: _T = None):
        '''
            Fill default_value if value is None
        '''
        self._data = [value if value is not None else self._defualt_value] * (self.width * self.height)
    
    def set(self, x: int, y: int, value: _T):
        self._data[y * self.width + x] = value
    def get(self, x: int, y: int) -> _T:
        return self._data[y * self.width + x]
    
    def convert(self, func: Callable[[_T], any]) -> 'Buffer':
        '''
            Create and convert a buffer\n
            `Don't modify` this buffer, just `create a new one`
        '''
        
        return Buffer(*self.size, self._defualt_value, (func(item) for item in self._data))        
    
    def resize(self, width: int, height: int) -> 'Buffer':
        if self.size == (width, height):
            return self.copy()
        
        buffer = Buffer(width, height, self._defualt_value)
        
        xk = self.width / width 
        yk = self.height / height
        
        for y in range(buffer.height):
            for x in range(buffer.width):
                buffer[x, y] = self[
                    min(max(math.floor(x*xk), 0), self.width-1), 
                    min(max(math.floor(y*yk), 0), self.height-1)
                ]
        
        return buffer
    
    @property
    def size(self) -> Vec2[int]:
        return self._size
    @property
    def width(self) -> int:
        return self.size.width
    @property
    def height(self) -> int:
        return self.size.height
    @property
    def data(self) -> tuple[_T]:
        return tuple(self._data)
    
    def __len__(self) -> int:
        return len(self._data)
    
    def __iter__(self):
        yield from self._data
                 
    def __getitem__(self, pos: tuple[int, int]) -> _T:
        return self.get(*pos)
    def __setitem__(self, pos: tuple[int, int], value: _T):
        self.set(*pos, value)

    @staticmethod
    def _raise_resize():
        raise RuntimeError('Size cannot be changed')
    
    def copy(self) -> 'Buffer':
        return Buffer(
            *self.size, self._defualt_value, self._data
        )
    
Buffer.empty = Buffer(0, 0, None)


from typing import Union, Iterable

from ..Classes import Vec3
from ..Config import Config

class Pixel:
    empty: 'Pixel' = None
    clearANSII = f'\033[0m'
    
    def __init__(
        self, 
        front_color: Union[Vec3[int], Iterable[int]] = None, 
        back_color: Union[Vec3[int], Iterable[int]] = None, 
        symbol: chr = None
    ):
        self.front_color = Vec3(*front_color) if isinstance(front_color, Iterable) else front_color
        self.back_color = Vec3(*back_color) if isinstance(back_color, Iterable) else back_color
        self.symbol = symbol if symbol is not None else ' '
        
        if isinstance(self.front_color, int):
            raise

    
    @staticmethod
    def fromRGB(br: int, bg: int, bb: int, symbol: chr = None) -> 'Pixel':
        return Pixel(back_color=Vec3(br, bg, bb), symbol=symbol)
    
    @staticmethod
    def changePixel(pixel: Union['Pixel', None], symbol: chr = None, front_color: Vec3 = None, back_color: Vec3 = None) -> Union['Pixel', None]:
        '''
            create a copy of the pixel and change it
        '''
        
        new_pixel = pixel.copy() if pixel is not None else Pixel.empty.copy()
        
        if symbol is not None:
            new_pixel.symbol = symbol
        
        if front_color is not None:
            new_pixel.front_color = front_color
        
        if back_color is not None:
            new_pixel.back_color = back_color    
        
        return new_pixel
    
    # WIP
    
    # @staticmethod
    # def mixStatic(col1: 'Pixel', col2: 'Pixel', alpha: float, symbol: chr = None, threshold: float = 0.005) -> 'Pixel':
    #     '''
    #         alpha: float = 0-1
    #         threshold: float = 0-1
    #     '''
        
    #     dvas = ((abs(col2.r - col1.r) + abs(col2.g - col1.g) + abs(col2.b - col1.b)) / 3) / 255
        
    #     if dvas < threshold:
    #         return Pixel(*col2.getRGB())
    #     else:
    #         return Pixel(
    #             round(col1.r * (1 - alpha) + col2.r * alpha), 
    #             round(col1.g * (1 - alpha) + col2.g * alpha), 
    #             round(col1.b * (1 - alpha) + col2.b * alpha),
    #             symbol if symbol is not None else col1.symbol
    #         )
    
    # def mix(self, col, alpha: float, symbol: chr = None, threshold: float = 0.005) -> 'Pixel':
    #     return Pixel.mixStatic(self, col, alpha, symbol, threshold)
    
    @property
    def ANSII(self) -> str:
        return f'{self.ANSIICol}{'⌂' if Config.DEBUG_SHOW_ANSIICOLOR_CHARS else ''}{self.symbol}'
    
    @property
    def ANSIICol(self) -> str:
        if self.back_color is None:
            if self.front_color is None:
                return Pixel.clearANSII
            else:
                return f'\033[38;2;{self.front_color.x};{self.front_color.y};{self.front_color.z};49m'
        else:
            if self.front_color is None:
                return f'\033[48;2;{self.back_color.x};{self.back_color.y};{self.back_color.z};39m'
            else:
                return f'\033[38;2;{self.front_color.x};{self.front_color.y};{self.front_color.z};48;2;{self.back_color.x};{self.back_color.y};{self.back_color.z}m'

    def getRGB(self) -> tuple[Vec3[int]]:
        '''
            return (front_color, back_color)
        '''
        return self.front_color, self.back_color

    @staticmethod
    def compareColors(pixel1: 'Pixel', pixel2: 'Pixel') -> bool:
        return pixel1.front_color == pixel2.front_color and pixel1.back_color == pixel2.back_color
    
    def getColors(self) -> tuple[Vec3, Vec3]:
        return (
            self.front_color,
            self.back_color
        )
    
    def __str__(self):
        return f'Pixel(f:{self.front_color};b:{self.back_color})'

    def copy(self) -> 'Pixel':
        return self.__class__(
            self.front_color,
            self.back_color,
            self.symbol
        )

Pixel.empty = Pixel()

class Pixels:
    f_white = Pixel((255, 255, 255))
    b_white = Pixel(None, (255, 255, 255))
    f_green = Pixel((0, 255, 0))
    b_green = Pixel(None, (0, 255, 0))
    f_gray = Pixel((125, 125, 125))
    b_gray = Pixel(None, (125, 125, 125))
    f_black = Pixel((0, 0, 0))
    b_black = Pixel(None, (0, 0, 0))
    f_red = Pixel((255, 0, 0))
    b_red = Pixel(None, (255, 0, 0))
    f_blue = Pixel((0, 0, 255))
    b_blue = Pixel(None, (0, 0, 255))
    f_yellow = Pixel((255, 255, 0))
    b_yellow = Pixel(None, (255, 255, 0))
    white_black = Pixel((255, 255, 255))
    black_white = Pixel((0, 0, 0), (255, 255, 255))



import asyncio
import os
import sys
from typing import Union
import importlib
import time

from .Config import Config
from .Threads import *
from .Classes.Event import Event
from .Log import Log
from .Managers import KeyboardManager as KeyM, WindowManager, Keys
from . import Func


class Core:   
    init_event: Event = Event()
    term_event: Event = Event()
    
    init_all_event = BaseThread.init_all_event
    
    GraphicThread = GraphicThread()
    SimulationThread = SimulationThread()
    InputThread = InputThread()

    _inited = False
    _tasks: list[asyncio.Task] = []
    
    @classmethod
    def start(cls):
        '''
            run async app
        '''
        try:
            if cls._inited is False:
                raise RuntimeError('Core was not initialized')
            
            for thread in BaseThread._threads.values():
                cls._tasks.append(Config.ASYNC_EVENT_LOOP.create_task(thread.run(), name=thread.name))
            
            Config.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
        
        except KeyboardInterrupt:
            Log.writeWarning('Emergency termination', cls)
            
        except:
            Log.writeError(cls)
            input('Press to continue...')
                
        finally:
            for task in cls._tasks:
                task.cancel()
            Config.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
            Config.ASYNC_EVENT_LOOP.stop()
            
    @classmethod
    def init(cls):
        if Config.CORE_MODIFY_WIN_REGEDIT and sys.platform == 'win32':
            import winreg
            
            try:
                reg_value = winreg.QueryValueEx(
                    winreg.OpenKey(
                        winreg.HKEY_CURRENT_USER, 
                        'Console', 
                        access=winreg.KEY_READ
                    ), 
                    'VirtualTerminalLevel'
                )
            except:
                reg_value = None
            
            if reg_value is None or reg_value[0] != 1:
                winreg.SetValueEx(
                    winreg.OpenKey(
                        winreg.HKEY_CURRENT_USER, 
                        'Console', 
                        access=winreg.KEY_SET_VALUE
                    ), 
                    'VirtualTerminalLevel',
                    0,
                    winreg.REG_DWORD,
                    1
                )
                
                input('Regedit has been modified, please restart the application\nPress to close...')
                exit()
                
        # KeyM.registerEvent('_close', Keys.CTRL_C)
        # KeyM.bind('_close', BaseThread.stopAll)
        
        cls.SimulationThread.delay = (1/Config.SPS) if Config.SPS > 0 else 0
        cls.GraphicThread.delay = (1/Config.FPS) if Config.FPS > 0 else 0
        
        cls.init_event.invoke()
        
        cls._inited = True
        Log.writeOk('Initialized', cls)
    
    @classmethod
    def term(cls):
        if cls._inited is False:
            raise RuntimeError('Core was not initialized')
        
        cls.init_event.invoke()
        
        cls._inited = False
        Log.writeOk('Terminated', cls)
    
    _dynamic_modules: dict[str, dict[str, any]] = {}
    @classmethod
    def addDynamicModule(cls, path: str, name: str):
        if path in cls._dynamic_modules:
            raise ValueError(f'Module "{path}" already exists')
        
        if not os.path.exists(path):
            raise ValueError(f'File "{path}" not exists')
        
        if Config.CORE_WRITE_WARNING_DYNAMIC_MODULE and len(cls._dynamic_modules) == 0:
            Log.writeWarning(
                '\nThis tool is unstable due to its features and may lead to errors\n' + 
                'It is strongly recommended to only change variables inside the module\n' + 
                'No complex logic', 
                cls
            )
            time.sleep(0.5)
        
        cls._dynamic_modules[path] = {
            'mtime': os.path.getmtime(path),
            'name': name,
            'module': importlib.import_module(name)
        }
        Log.writeOk(f'module "{path}" added', cls)
        
    @classmethod
    def popDynamicModule(cls, path: str):
        if cls._dynamic_modules.pop(path, None) is not None:
            Log.writeOk(f'module "{path}" removed', cls)
    
    @classmethod
    def checkDynamicModules(cls):
        for path, data in cls._dynamic_modules.items():
            if not os.path.exists(path):
                continue
            
            os.path.dirname
            
            last_mtime = os.path.getmtime(path)
            if data['mtime'] == last_mtime:
                continue
            
            data['module'] = importlib.reload(data['module'])
            data['mtime'] = last_mtime
            
            Log.writeOk(f'module "{path}" updated')
            
    

             


from typing import Union, Iterable
import os
import playsound
import asyncio

from ..Log import Log

class SoundManager:
    # name: path
    _sounds: dict[str, str] = {}

    @classmethod
    def registerSound(cls, name: str, path: str, rewrite: bool = False):
        if not os.path.exists(path):
            raise ValueError(f'File "path" not found')
        
        if name in cls._sounds and not rewrite:
            Log.writeWarning(f'Sound "{name}" already registered', cls)
            
        cls._sounds[name] = path
        
    @classmethod
    def play(cls, name: str):
        try:
            playsound.playsound(cls._sounds[name], False)
        except:
            Log.writeError(cls)
        



from typing import Union, Iterable

from ..Classes import Vec2, Vec3, Vec4, Event, Buffer, Orientation
from .Pixel import Pixel, Pixels
from .. import Converter


class BaseGraphicObject:
    def __init__(
        self,
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        can_be_moved: bool = True,
        *args, **kwargs
        ):
        
        ''' events '''
        self.__resize_event = Event(
            self.setFlagRefresh
        )
        self.__change_clear_pixel_event = Event(
            self.setFlagRefresh
        )
        self.__set_refresh_event = Event()
        
        ''' size and pos '''
        self._offset_pos = Converter.toVec3(offset_pos)
        self._size = Converter.toVec2(size)
        self._padding: Vec4[int] = Converter.toVec4(padding)
        self._min_size = Converter.toVec2(min_size, Vec2(None, None), True)
        self._max_size = Converter.toVec2(max_size, Vec2(None, None), True)
        self._can_be_moved = can_be_moved
        
        ''' buffers '''
        self._buffer: Buffer[Pixel] = None
        
        ''' pixels '''
        self._clear_pixel = Converter.toPixel(clear_pixel)
        
        ''' flags '''
        self._flag_refresh = True
        
        ''' other '''
        self._name = name
    
    async def refresh(self):
        self._buffer = Buffer(
            self.width + self.padding.horizontal,
            self.height + self.padding.vertical,
            self.clear_pixel
        )
        
        self._flag_refresh = False
    
    async def awaitingRefresh(self) -> bool:
        return False
    
    async def render(self) -> Buffer[Pixel]:
        if self._flag_refresh:
            await self.refresh()
        return self._buffer
    
    def setFlagRefresh(self):
        self._flag_refresh = True
        self.set_refresh_event.invoke()
    
    def setOffsetPos(self, value: Vec3[int]):
        self._offset_pos = value
    def getOffsetPos(self) -> Vec3[int]:
        return self._offset_pos
    @property
    def offset_pos(self) -> Vec3[int]:
        return self.getOffsetPos()
    @offset_pos.setter
    def offset_pos(self, value: Vec3[int]):
        self.setOffsetPos(value)
    @property
    def offset_x(self) -> int:
        return self.offset_pos.x
    @offset_x.setter
    def offset_x(self, value: int):
        self.offset_pos.x = value
    @property
    def offset_y(self) -> int:
        return self.offset_pos.y
    @offset_y.setter
    def offset_y(self, value: int):
        self.offset_pos.y = value
    @property
    def offset_z(self) -> int:
        return self.offset_pos.z
    @offset_z.setter
    def offset_z(self, value: int):
        self.offset_pos.z = value
    
    def setSize(self, value: Vec2[int]):
        self._size = value
        self.resize_event.invoke()
        value.change_event.add(
            self.resize_event.invoke
        )
    def getSize(self) -> Vec2[int]:
        return Vec2(
            max(
                max(
                    self._size.width,  
                    self._min_size.width if self._min_size.width is not None else 0
                ), 
                min(
                    self._size.width,
                    self._max_size.width if self._max_size.width is not None else self._size.width
                )
            ),
            max(
                max(
                    self._size.height, 
                    self._min_size.height if self._min_size.height is not None else 0
                ), 
                min(
                    self._size.height,
                    self._max_size.height if self._max_size.height is not None else self._size.height
                )
            )
        )
    @property
    def size(self) -> Vec2[int]:
        return self.getSize()
    @size.setter
    def size(self, value: Vec2[int]):
        self.setSize(value)
    @property
    def width(self) -> int:
        return self.size.width
    @width.setter
    def width(self, value: int):
        self._size.width = value
    @property
    def height(self) -> int:
        return self.size.height
    @height.setter
    def height(self, value: int):
        self._size.height = value
        
    @property
    def min_size(self) -> Vec2[int]:
        return self._min_size
    @size.setter
    def min_size(self, value: Vec2[int]):
        self._min_size = value
        self.setFlagRefresh()
        
    @property
    def max_size(self) -> Vec2[int]:
        return self._max_size
    @size.setter
    def max_size(self, value: Vec2[int]):
        self._max_size = value
        self.setFlagRefresh()
    
    @property
    def name(self) -> Union[str, None]:
        return self._name
    
    def getClearPixel(self) -> Union[Pixel, None]:
        return self._clear_pixel
    
    def setClearPixel(self, value: Union[Pixel, None]):
        self._clear_pixel = value
        self.change_clear_pixel_event.invoke()
    
    @property
    def clear_pixel(self) -> Union[Pixel, None]:
        return self.getClearPixel()
    @clear_pixel.setter
    def clear_pixel(self, value: Union[Pixel, None]):
        self.setClearPixel(value)
    
    @property
    def resize_event(self) -> Event:
        return self.__resize_event
    @property
    def change_clear_pixel_event(self) -> Event:
        return self.__change_clear_pixel_event
    @property
    def set_refresh_event(self) -> Event:
        return self.__set_refresh_event
    
    def setPadding(self, value: Vec4[int]):
        self._padding = value
        self.__resize_event.invoke()
        value.change_event.add(
            self.__resize_event.invoke
        )
    def getPadding(self) -> Vec4[int]:
        return self._padding
    @property
    def padding(self) -> Vec4[int]:
        '''
            `up`: 0 | x\n
            `bottom` 1 | y\n
            `left` 2 | z\n
            `right` 3 | w
        '''
        return self.getPadding()
    @padding.setter
    def padding(self, value: Vec4[int]):
        self.setPadding(value)
        
    @property
    def can_be_moved(self) -> bool:
        return self._can_be_moved
    @can_be_moved.setter
    def can_be_moved(self, value: bool):
        self._can_be_moved = value


class BaseGraphicContainerObject(BaseGraphicObject):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        objects: Union[Iterable[BaseGraphicObject], BaseGraphicObject] = [], 
        direction: Union[Orientation, None] = Orientation.horizontal,
        gap: int = 0,
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        super().__init__(
            size=size,
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        
        ''' size and pos '''
        self._auto_size = auto_size
        
        ''' events '''
        self.__add_object_event = Event(
            self.setFlagRefresh
        )
        
        ''' objects '''
        self._objects: list['BaseGraphicObject'] = []
        for object in Converter.toListObjects(objects):
            self.addObject(object)
        self._direction: Orientation = Converter.toOrientation(direction)
        self._gap: int = gap
        
        ''' buffers '''
        self._objects_buffer: Buffer[Pixel] = Buffer.empty
    
    async def refresh(self):
        objects = [
            (
                object.offset_x,
                object.offset_y,
                await object.render(),
                object.can_be_moved
            )
            for object in self._objects
        ]
        
        if self._direction is None:
            await super().refresh()
            
            self._objects_buffer = Buffer(
                max(self.width - self.padding.horizontal, 0),
                max(self.height - self.padding.vertical, 0),
                self.clear_pixel
            )
            
            for object_data in objects:
                self._objects_buffer.paste(*object_data[:3])
                
            
        else:
            object_buffer_width = object_buffer_height = 0
            
            for object in self._objects:
                if object.can_be_moved is False:
                    continue
                
                match self._direction:
                    case Orientation.vertical:
                        object_buffer_width = max(object.width + object.padding.horizontal, object_buffer_width)
                    case _:
                        object_buffer_width += object.width + object.padding.horizontal + self.gap
                
                match self._direction:
                    case Orientation.horizontal:
                        object_buffer_height = max(object.height + object.padding.vertical, object_buffer_height)
                    case _:
                        object_buffer_height += object.height + object.padding.vertical + self.gap
            
            if self._direction is Orientation.horizontal:
                object_buffer_width -= 1
            else:
                object_buffer_height -= 1
            
            self._objects_buffer = Buffer(
                object_buffer_width,
                object_buffer_height,
                self.clear_pixel
            )
            
            self.size = Vec2(
                self._objects_buffer.width,
                self._objects_buffer.height
            )
            
            await super().refresh()
            
            x_indent = y_indent = 0
            for object in objects:
                if object[3] is True: # object can be moved
                    self._objects_buffer.paste(
                        x_indent, y_indent,
                        object[2]
                    )
                    
                    if self._direction is Orientation.horizontal:
                        x_indent+=object[2].width + self.gap
                    else:
                        y_indent+=object[2].height + self.gap
                else:
                    self._objects_buffer.paste(*object[:3])
                
        self._buffer.paste(
            self.padding.left, self.padding.top,
            self._objects_buffer
        )
    
    async def render(self):
        for object in self._objects:
            if await object.awaitingRefresh():
                self.setFlagRefresh()
                break
        
        return await super().render()
    
    def addObject(self, object: BaseGraphicObject):
        self._objects.append(
            object
        )
        object.set_refresh_event.add(self.setFlagRefresh)
        self.add_object_event.invoke()
    
    @property
    def add_object_event(self) -> Event:
        return self.__add_object_event
    
    @property
    def gap(self) -> int:
        return self._gap
    @gap.setter
    def gap(self, value: int):
        self._gap = value
    
    def __iter__(self):
        yield from self._objects
    
    def __str__(self, **kwargs):
        kwargs.update({
            "name": self._name,
            "size": self._size,
            "offset_pos": self._offset_pos
        })
        return f'{self.__class__.__name__}({' '.join([f'{key}:{value}' for key, value in kwargs.items()])})'



from typing import Union, Iterable
import numpy as np 

from ..BaseGraphicObject import BaseGraphicObject, BaseGraphicContainerObject
from ..Pixel import Pixel, Pixels
from ..Drawer import Drawer
from ..Widgets.Widget import Widget
from ..Widgets.InteractiveWidget import InteractiveWidget
from ...Classes import Buffer, Event, Vec2, Vec3, Vec4, Keys, Orientation
from ...Log import Log

from ... import Converter
from ... import Func

class Window(BaseGraphicContainerObject):    
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = [], 
        direction: Union[Orientation, None] = Orientation.horizontal,
        gap: int = 0,
        can_be_moved: bool = True,
        frame: bool = False,
        frame_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        *args, **kwargs
    ):

        
        super().__init__(
            size=size, 
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            auto_size=auto_size, 
            offset_pos=offset_pos, 
            clear_pixel=Func.choisePixel(clear_pixel, Pixel.empty), 
            name=name, 
            objects=widgets, 
            direction=direction,
            gap=gap,
            can_be_moved=can_be_moved,
        *args, **kwargs)
        
        ''' events '''
        self._open_event = Event()
        self._close_event = Event()
        self._change_frame_pixel_event = Event()
        
        ''' pixels ''' 
        self._frame_pixel = Converter.toPixel(frame_pixel)
        
        ''' interact_objects '''
        self._select_index: int = 0
        self._interact_objects: list[InteractiveWidget] = []
        self.updateInteractWidgets()
        self.add_object_event.add(self.updateInteractWidgets)
        
        ''' other '''
        self._frame = frame
        
    async def refresh(self):
        await super().refresh()
        
        if self.frame:
            frame_pixel: Pixel = Func.choisePixel(
                self.frame_pixel, 
                self.clear_pixel
            )
            
            self._buffer.paste(
                0, 0,
                Drawer.frame(
                    self.width + self.padding.horizontal,
                    self.height + self.padding.vertical,
                    frame_pixel.front_color, 
                    frame_pixel.back_color
                )
            )
        
    def getPadding(self):
        return super().getPadding() + (
            Vec4(1, 1, 1, 1) if self.frame else Vec4(0, 0, 0, 0)
        )
    
    def updateInteractWidgets(self):
        def _f(container_object: BaseGraphicContainerObject) -> list[InteractiveWidget]:
            widgets = []
            for object in container_object:
                if issubclass(object.__class__, BaseGraphicContainerObject):
                    widgets += _f(object)
                
                if issubclass(object.__class__, InteractiveWidget):
                    widgets.append(object)
            return widgets
        
        self._interact_objects = _f(self._objects)
        self.selectWidget(0) 
    
    
    def _normalizeSelectIndex(self):
        self._select_index = Func.normalizeIndex(self._select_index, len(self._interact_objects))
    
    def getSelectedWidget(self) -> Union[InteractiveWidget, None]:
        if len(self._interact_objects) == 0:
            return None
        self._normalizeSelectIndex()
        return self._interact_objects[self._select_index]
    
    def selectWidget(self, index: int):
        if len(self._interact_objects) == 0:
            return
        
        previous_widget = self.getSelectedWidget()
        if previous_widget is not None:
            previous_widget.selected = False
            
        self._select_index = index
        self._normalizeSelectIndex()
        
        next_widget = self.getSelectedWidget()
        if next_widget is not None:
            next_widget.selected = True
    
    def selectNext(self):
        self.selectWidget(self._select_index + 1)
    
    def selectPrevious(self):
        self.selectWidget(self._select_index - 1)
    
    def inputKey(self, key: str) -> bool:
        match key:
            case Keys.UP:
                self.selectPrevious()
                
            case Keys.DOWN:
                self.selectNext()
                
            case _:
                widget = self.getSelectedWidget()
                if widget is not None:
                    return widget.inputKey(key)
                
                return False
        return True
    
    @property
    def open_event(self) -> Event:
        return self._open_event
    @property
    def close_event(self) -> Event:
        return self._close_event
    
    def setFrame(self, value: bool):
        self._frame = value
        self.setFlagRefresh()
    @property
    def frame(self) -> bool:
        return self._frame
    @frame.setter
    def frame(self, value: bool):
        self.setFrame(value)
    
    @property
    def frame_pixel(self) -> Union[Pixel, None]:
        return self._frame_pixel
    @frame_pixel.setter
    def frame_pixel(self, value):
        self._frame_pixel = value
        self.setFlagRefresh()
        self.change_frame_pixel_event.invoke()

    @property
    def change_frame_pixel_event(self) -> Event:
        return self._change_frame_pixel_event



from setuptools import setup, find_packages

def readme():
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()

setup(
    name='FloriaConsoleGUI',
    version='1.1.1',
    author='FloriaProduction',
    author_email='FloriaProduction@yandex.ru',
    description='Framework for console GUI apps',
    long_description=readme(),
    long_description_content_type='text/markdown',
    license='Apache-2.0',
    project_urls={
        "Documentation": 'https://FloriaConsoleGUI.github.io',
        "Source": 'https://github.com/FloriaProduction/FloriaConsoleGUI',
    },
    packages=find_packages(),
    install_requires=['readchar', 'pillow', 'playsound==1.2.2'],
    classifiers=[
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: Apache Software License',
        'Environment :: Console',
        "Intended Audience :: Developers",
        "Typing :: Typed"
    ],
    keywords='python console gui ui ux',
    zip_safe=True,
    python_requires='>=3.12'
)


from typing import Union, Iterable

from .Container import Container
from .Widget import Widget
from ..Pixel import Pixel, Pixels
from ...Classes import Vec2, Vec3, Vec4, Orientation
from ..Drawer import Drawer
from ... import Func, Converter

class Frame(Container):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        max_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        frame_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None, 
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None,
        direction: Union[Orientation, None] = Orientation.horizontal,
        gap: int = 0,
        can_be_moved: bool = True,
        *args, **kwargs
    ):
        super().__init__(
            size=size,
            min_size=min_size,
            max_size=max_size,
            padding=padding,
            auto_size=auto_size, 
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            widgets=widgets, 
            direction=direction,
            gap=gap,
            can_be_moved=can_be_moved,
            *args, **kwargs
        )
        
        self._frame_pixel = Converter.toPixel(frame_pixel)
    
    def getPadding(self):
        return super().getPadding() + Vec4(1, 1, 1, 1)
        
    async def refresh(self):
        await super().refresh()
        
        frame_pixel: Pixel = Func.choisePixel(
            self.frame_pixel, 
            self.clear_pixel, 
            default=Pixels.white_black
        )
        
        self._buffer.paste(
            0, 0,
            Drawer.frame(
                *self.size, 
                frame_pixel.front_color,
                frame_pixel.back_color
            ),
            func=Drawer.mergeFramePixels
        )
    
    @property
    def frame_pixel(self) -> Pixel:
        return self._frame_pixel
    @frame_pixel.setter
    def frame_pixel(self, value: Union[Pixel, None]):
        self._frame_pixel = value
    
        




from .BaseGraphicObject import BaseGraphicObject, BaseGraphicContainerObject
from .Pixel import Pixel, Pixels
from .Drawer import Drawer
from .Animation import Animation

__all__ = [
    'Pixel', 'Pixels'
    'Windows', 
    'Widgets',
    'Drawer',
    'Animation'
]



from typing import Union
from traceback import format_exc
from time import sleep
from .Config import Config

class Log:
    _min_state = _min_name = 0

    @classmethod
    def write(
        cls, 
        message: str, 
        state: str, 
        name: Union[str, any, None] = None, 
        message_color: str = '\033[38;2;125;125;125;49m', 
        state_color: str = '\033[38;2;125;125;125;49m', 
        name_color: str = '\033[38;2;125;125;125;49m',
        capitalize_message: bool = True
    ):
        cls._min_state = max(cls._min_state, len(state))
        
        if not isinstance(name, str) and name is not None:
            if isinstance(name, type):
                name = name.__name__
            elif isinstance(name, object):
                name = name.__class__.__name__
        cls._min_name = max(cls._min_name, len(name)) if name is not None else cls._min_name

        if capitalize_message:
            message = message[0].capitalize() + message[1:]
        
        print(
            '\033[0m' + 
            f'{state_color}{f'[{state.upper()}]'.ljust(cls._min_state + 2)}  ' + 
            (f'{name_color}{f'[{name}]'.ljust(cls._min_name + 2)}  ' if name is not None else '') + 
            f'{message_color}{message}\033[0m'
        )
    
    @classmethod
    def writeOk(cls, message: str, name: Union[str, None] = None):
        cls.write(
            message, 
            'ok', 
            name, 
            state_color='\033[32;49m', 
            message_color='\033[37;49m'
        )
    
    @classmethod
    def writeNotice(cls, message: str, name: Union[str, None] = None):
        cls.write(
            message, 
            'notice', 
            name
        )
    
    @classmethod
    def writeError(cls, name: Union[str, None] = None, message: Union[str, None] = None):
        cls.write(
            message if message is not None else format_exc(), 
            'error', 
            name, 
            state_color='\033[31;49m', 
            name_color='\033[37;49m', 
            message_color='\033[37;49m', 
            capitalize_message=False
        )
        sleep(Config.LOG_ERROR_DELAY)

    @classmethod
    def writeWarning(cls, message: Union[str, None] = None, name: Union[str, None] = None):
        cls.write(
            message if message is not None else format_exc(), 
            'warning', 
            name, 
            state_color='\033[33;49m', 
            message_color='\033[37;49m'
        )





from .Widget import Widget
from .Frame import Frame
from .Container import Container
from .ProgressBar import ProgressBar
from .Scroll import Scroll
from .InteractiveWidget import InteractiveWidget
from .Media import Media
from .Button import Button
from .CheckBox import CheckBox
from .TextBox import TextBox
from .Label import Label


from typing import Union, Iterable

from ..Classes import Buffer, Vec3
from .Pixel import Pixel
from ..Config import Config

# ┘ └ │ ├ ┬
# ┐ ┌ ─ ┴ ┤ ┼
        
class FramePixel(Pixel):
    pass

class Drawer:
    _cache: dict[str, Buffer[Pixel]] = {}
    
    @classmethod
    def _putToCache(cls, key: str, value: Buffer[Pixel]):
        cls._cache[key] = value
        if len(cls._cache) > Config.DRAWER_MAX_SIZE_CACHE:
            cls._cache.pop(tuple(cls._cache.keys())[0])
    
    @classmethod
    def _genKey(cls, **kwargs) -> str:
        return '_'.join([f'{key}:{value}' for key, value in kwargs.items()])
    
    @classmethod
    def frame(cls, width: int, height: int, front_color: Union[Vec3, None], back_color: Union[Vec3, None]) -> Buffer[Pixel]:        
        if width == 0 or height == 0:
            return Buffer.empty
        
        key = cls._genKey(func='frame', width=width, height=height, back_color=back_color, front_color=front_color)
        
        if key not in cls._cache:
            buffer: Buffer[Pixel] = Buffer(width, height, None)
            
            pixel = FramePixel(front_color, back_color)
            
            for x in range(1, width-1):
                buffer[x, 0] = buffer[x, height-1] = Pixel.changePixel(pixel, '─')
                
            for y in range(1, height-1):
                buffer[0, y] = buffer[width-1, y] = Pixel.changePixel(pixel, '│')
            
            buffer[0, 0]              = Pixel.changePixel(pixel, '┌')
            buffer[0, height-1]       = Pixel.changePixel(pixel, '└')
            buffer[width-1, 0]        = Pixel.changePixel(pixel, '┐')
            buffer[width-1, height-1] = Pixel.changePixel(pixel, '┘')
            
            cls._putToCache(key, buffer)
            
        return cls._cache[key]
    
    @classmethod
    def mergeFramePixels(
        cls, 
        x_parent: int, 
        y_parent: int, 
        width_parent: int, 
        height_parent: int, 
        x_child: int, 
        y_child: int, 
        width_child: int, 
        height_child: int, 
        pixel1: Union[FramePixel, None], 
        pixel2: Union[FramePixel, None]
    ) -> FramePixel:       
        def inArr(*args) -> bool:
            for symbol in args:
                if symbol not in arr:
                    return False
            return True
        
        if pixel2 is None:
            return pixel1
               
        if not (isinstance(pixel1, FramePixel) and isinstance(pixel2, FramePixel)):
            return pixel2
        
        new_symbol = pixel2.symbol
        arr = (pixel1.symbol, pixel2.symbol)
        
        # ┘ └ │ ├ ┬
        # ┐ ┌ ─ ┴ ┤ ┼
        
        if arr[0] == arr[1]:
            new_symbol = pixel2.symbol
        elif inArr('─', '│') or inArr('┼') or \
            inArr('┐', '└') or inArr('┘', '┌') or \
            inArr('├', '┤') or inArr('┴', '┬') or \
            inArr('├', '┘') or inArr('├', '┐') or \
            inArr('┤', '└') or inArr('┤', '┌') or \
            inArr('┬', '┘') or inArr('┴', '┐') or \
            inArr('┬', '└') or inArr('┴', '┌'):
            new_symbol = '┼'
        elif inArr('┌', '└') or inArr('├') or \
            inArr('│', '└') or inArr('│', '┌'):
            new_symbol = '├'
        elif inArr('┐', '┘') or inArr('┤') or \
            inArr('│', '┘') or inArr('│', '┐'):
            new_symbol = '┤'
        elif inArr('┘', '└') or inArr('┴') or \
            inArr('─', '└') or inArr('─', '┘'):
            new_symbol = '┴'
        elif inArr('┌', '┐') or inArr('┬') or \
            inArr('─', '┌') or inArr('─', '┐'):
            new_symbol = '┬'
        
        return Pixel.changePixel(pixel2, symbol=new_symbol)
    
    
    _text_buffers_cache: dict[str, Buffer[Pixel]] = {}
    @classmethod
    async def renderTextBuffer(cls, text: Union[str, Iterable[str]], text_pixel: Pixel = None) -> Buffer[Pixel]:
        name = f'{text}_{text_pixel}'
        
        if name not in cls._text_buffers_cache:
            lines = text.split('\n') if isinstance(text, str) else text
            lines_mod = [
                *[line + Config.NEW_LINE_SYMBOL for line in lines[:-1]], 
                lines[-1]
            ]
            buffer = Buffer(
                max(map(len, lines_mod)),
                len(lines_mod),
                text_pixel
            )
            for y in range(buffer.height):
                for x in range(min(len(lines_mod[y]), buffer.width)):
                    buffer.set(
                        x, y, 
                        Pixel.changePixel(buffer[x, y], symbol=lines_mod[y][x])
                    )
            cls._text_buffers_cache[name] = buffer
        return cls._text_buffers_cache[name].copy()